/*=====================================================================*/
/*    serrano/prgm/project/bigloo/fthread/src/bglthread.c              */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Fri Feb 22 12:12:04 2002                          */
/*    Last change :  Wed Aug  7 11:13:03 2002 (serrano)                */
/*    Copyright   :  2002 Manuel Serrano                               */
/*    -------------------------------------------------------------    */
/*    C utilities for native Bigloo fair threads implementation.       */
/*=====================================================================*/
#include <pthread.h>
#include <sched.h>
#include <stdlib.h>
#include <gc.h>

#define GC_PRIVATE_H
#include <bglthread.h>

/*---------------------------------------------------------------------*/
/*    Runtime system                                                   */
/*---------------------------------------------------------------------*/
extern bglthread_rts_init();
extern char *strerror();

/*---------------------------------------------------------------------*/
/*    Asynchronous threads                                             */
/*---------------------------------------------------------------------*/
extern pthread_mutex_t async_lock;
extern pthread_cond_t async_cv;

/*---------------------------------------------------------------------*/
/*    Global token                                                     */
/*---------------------------------------------------------------------*/
pthread_key_t bglkey;

static bglthread_t token;

static pthread_mutex_t first_lock;
static pthread_cond_t first_cv;

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    bglthread_init ...                                               */
/*---------------------------------------------------------------------*/
static void
bglthread_init() {
   static int bglthread_initp = 0;
   
   if( !bglthread_initp ) {
      bglthread_initp = 1;
      token = 0L;

      pthread_key_create( &bglkey, 0L );

      pthread_mutex_init( &first_lock, 0L );
      pthread_cond_init( &first_cv, 0L );

      pthread_mutex_init( &async_lock, 0L );
      pthread_cond_init( &async_cv, 0L );
      
      bglthread_rts_init();
   }
}

/*---------------------------------------------------------------------*/
/*    bglthread_t                                                      */
/*    bglthread_new ...                                                */
/*---------------------------------------------------------------------*/
bglthread_t
bglthread_new( obj_t proc ) {
   bglthread_t thread = (bglthread_t)GC_MALLOC( sizeof( struct bglthread ) );
   
   bglthread_init();
   
   thread->thunk = proc;
   thread->started = 0;
   
   thread->exitd_top = BFALSE;
   thread->befored_top = 0L;
   
   thread->current_output_port = bgl_get_current_output_port();
   thread->current_error_port = bgl_get_current_error_port();
   thread->current_input_port = bgl_get_current_input_port();

   thread->mvalues_number = 1;
   thread->mvalues_values;
   thread->mvalues_values [ 0 ] = BUNSPEC;
   thread->mvalues_values [ 1 ] = BUNSPEC;
   thread->mvalues_values [ 2 ] = BUNSPEC;
   thread->mvalues_values [ 3 ] = BUNSPEC;
   
   pthread_mutex_init( &(thread->lock), 0L );
   pthread_cond_init( &(thread->cv), 0L );

   return thread;
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bglthread_setup ...                                              */
/*---------------------------------------------------------------------*/
void
bglthread_setup( bglthread_t thread, obj_t obj ) {
   thread->bglobj = obj;
}

/*---------------------------------------------------------------------*/
/*    bglthread_t                                                      */
/*    pthread_current_thread ...                                       */
/*---------------------------------------------------------------------*/
bglthread_t
pthread_current_thread() {
   void *obj = pthread_getspecific( bglkey );
   return obj ? (bglthread_t)obj : 0L;
}

/*---------------------------------------------------------------------*/
/*    static void *                                                    */
/*    bglthread_run ...                                                */
/*---------------------------------------------------------------------*/
static void *
bglthread_run( void *arg ) {
   bglthread_t thread = (bglthread_t)arg;
   bglthread_init();

   thread->started = 1;
   thread->stack_bottom = (char *)&arg;
   
   /* we have to store in the thread, the pointer to the bglthread */
   pthread_setspecific( bglkey, arg );
   
/*    printf( "bglthread_run: bglthread=%p thread=%p\n", thread, &thread->pthread ); */
   while( token != thread ) {
      pthread_mutex_lock( &(thread->lock) );
      pthread_cond_wait( &(thread->cv), &(thread->lock) );
      pthread_mutex_unlock( &(thread->lock) );
   }

   PROCEDURE_ENTRY( thread->thunk )( thread->thunk, BEOA );
   return BUNSPEC;
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bglthread_start ...                                              */
/*---------------------------------------------------------------------*/
void
bglthread_start( bglthread_t thread ) {
   if( pthread_create( &(thread->pthread), 0L, bglthread_run, thread ) )
      FAILURE( string_to_bstring( "thread-start!" ),
	       string_to_bstring( "Cannot start thread" ),
	       string_to_bstring( strerror( errno ) ) );
   if( pthread_detach( thread->pthread ) )
      FAILURE( string_to_bstring( "thread-start!" ),
	       string_to_bstring( "Cannot start thread" ),
	       string_to_bstring( strerror( errno ) ) );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bglthread_await ...                                              */
/*---------------------------------------------------------------------*/
void
bglthread_await( bglthread_t t, bglthread_t s ) {
   pthread_cond_wait( &async_cv, &async_lock );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bglthread_switch_to ...                                          */
/*---------------------------------------------------------------------*/
void
bglthread_switch_to( bglthread_t this, bglthread_t next ) {
   while( !next->started ) sched_yield();

   pthread_mutex_lock( &(this->lock) );
   pthread_mutex_lock( &(next->lock) );
   token = next;
   pthread_cond_signal( &(next->cv) );
   pthread_mutex_unlock( &(next->lock) );
   while( token != this ) {
      pthread_cond_wait( &(this->cv), &(this->lock) );
   }
   pthread_mutex_unlock( &(this->lock) );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bglthread_switch_in ...                                          */
/*---------------------------------------------------------------------*/
void
bglthread_switch_in( bglthread_t next ) {
   pthread_mutex_t *this_lock;
   pthread_cond_t *this_cv;
   bglthread_t this;
   bglthread_init();

   while( !next->started ) sched_yield();
   this =  pthread_current_thread();

   next->parent = this;

   if( this ) {
      this_lock = &(this->lock);
      this_cv = &(this->cv);
   } else {
      this_lock = &first_lock;
      this_cv = &first_cv;
   }

   pthread_mutex_lock( this_lock );
   pthread_mutex_lock( &(next->lock) );
   token = next;
   pthread_cond_signal( &(next->cv) );
   pthread_mutex_unlock( &(next->lock) );
   
   while( token != this ) pthread_cond_wait( this_cv, this_lock );
   
   pthread_mutex_unlock( this_lock );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bglthread_switch_out ...                                         */
/*---------------------------------------------------------------------*/
void
bglthread_switch_out( bglthread_t this ) {
   pthread_mutex_t *parent_lock;
   pthread_cond_t *parent_cv;

/*    printf( "switch_out: parent=%p  this=%p\n", this->parent, this ); */
   if( this->parent ) {
      parent_lock = &(this->parent->lock);
      parent_cv = &(this->parent->cv);
   } else {
      parent_lock = &first_lock;
      parent_cv = &first_cv;
   }
   
   token = 0L;
   pthread_mutex_lock( parent_lock );
   pthread_cond_signal( parent_cv );
   pthread_mutex_unlock( parent_lock );

   pthread_mutex_lock( &(this->lock) );
   while( token != this ) pthread_cond_wait( &(this->cv), &(this->lock) );
   pthread_mutex_unlock( &(this->lock) );
}
      
   
